#ifndef OPM_MAKE_EDGE_CONFORMAL_HEADER
#define OPM_MAKE_EDGE_CONFORMAL_HEADER

#include <opm/grid/cpgpreprocess/preprocess.h>

#ifdef __cplusplus
extern "C" {
#endif // __cplusplus

int make_edge_conformal(struct processed_grid* grid);

#ifdef __cplusplus
}
#endif // __cplusplus

#endif // OPM_MAKE_EDGE_CONFORMAL_HEADER
