// -*- mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=2 sw=2 sts=2:
#ifndef DUNE_POLYHEDRALGRID_DECLARATION_HH
#define DUNE_POLYHEDRALGRID_DECLARATION_HH

namespace Dune
{

  template< int dim, int dimworld, typename coord_t = double >
  class PolyhedralGrid;

} // namespace Dune

#endif // #ifndef DUNE_POLYHEDRALGRID_DECLARATION_HH
